{-# LANGUAGE TypeOperators,
             MultiParamTypeClasses, 
             FlexibleInstances,
             FlexibleContexts,
             UndecidableInstances,
             GADTs #-}

module Compiler where

import SourceLanguage
import TargetLanguage
import Graph
import Fixpoint
import HFixpoint
import Prelude hiding (EQ, LT, rem)


infixr 0  |>
(|>) :: (GraphT g v -> GraphT g v) -> GraphT g v -> GraphT g v
(|>) = ($)
                  
class (HFunctor f) => CompAlgG f g where
  compAlgG :: f (K (GraphT g v -> GraphT g v)) l -> GraphT g v -> GraphT g v
  
instance (CompAlgG f g, CompAlgG h g) => CompAlgG (f ::+ h) g where
  compAlgG (HInl x)     = compAlgG x
  compAlgG (HInr y)     = compAlgG y
  
instance (ARITH :<: g) => CompAlgG Arith g where
  compAlgG (Val n)    c = pushG n |> c
  compAlgG (Add (K x) (K y))  c = x |> y |> addG |> c
  compAlgG (Mul (K x) (K y))  c = x |> y |> mulG |> c
  
instance (EXCEPT :<: g) => CompAlgG Except g where
  compAlgG (Throw)     _ = throwG
  compAlgG (Catch (K x) (K h)) c = letx c (\v -> markG (h |> Var v) |> x |> unmarkG |> Var v)

instance CompAlgG Seq g where
  compAlgG (Seq (K x) (K y)) c = x |> y |> c

instance (COND :<: g) => CompAlgG If g where
  compAlgG (If (K b) (K p1) (K p2)) c = letx c (\ v -> b |> jpcG (p1 |>  Var v) |> p2 |> Var v)
  
instance (COND :<: g) => CompAlgG While g where
  compAlgG (While (K b) (K lb)) c = mu (\ v -> b |> jpcG (lb |> Var v) |> c)

instance (STATE :<: g) => CompAlgG State g where
  compAlgG (Get v) c   = getG v |> c
  compAlgG (Set v (K e)) c = e |> setG v |> c

instance (BOOL :<: g) => CompAlgG Boolean g where
    compAlgG (BVal b) c = pushbG b |> c
    compAlgG (And (K x) (K y)) c = x |> y |> andG |> c
    compAlgG (Or (K x) (K y)) c = x |> y |> orG |> c

instance (COMP :<: g) => CompAlgG Comp g where
    compAlgG (Equ (K x) (K y)) c = x |> y |> eqG |> c
    compAlgG (Lt (K x) (K y)) c = x |> y |> ltG |> c


comp :: (HFunctor f, CompAlgG f g, HALT :<: g) => HFix f i -> Graph g
comp e = Graph (hfold (K . compAlgG) e `unK` haltG)
