
-- This is a variant of the Compiler module that implements exceptions
-- differently (with the handler code at the throw instruction). The
-- virtual machine implementation is amended accordingly.


{-# LANGUAGE TypeOperators, MultiParamTypeClasses, 
             FlexibleInstances, DeriveFunctor, 
             FlexibleContexts, UndecidableInstances,
             NoMonomorphismRestriction #-}

module CompilerCont where

import Compiler
import Fixpoint
import Graph
import HFixpoint
import SourceLanguage
import TargetLanguage hiding (EXCEPT (..), markG, unmarkG, throwG)
import VirtualMachine

-----------------------------------
-- Syntax of the Target Language --
-----------------------------------


data EXCEPT e = THROW e | UNMARK e | MARK e deriving (Functor, Show)


  
markG   = \c -> injectG (MARK c)
unmarkG = \  c -> injectG (UNMARK c)
throwG  h = injectG (THROW h)

    
--------------
-- Compiler --
--------------


                  
class CompAlgG' f g where
  compAlgG' :: f (K (GraphT g v -> GraphT g v -> GraphT g v)) l
          -> GraphT g v -> GraphT g v -> GraphT g v
  
comp' :: (HFunctor f, CompAlgG' f g, HALT :<: g) => HFix f i -> Graph g
comp' e = Graph (unK (hfold (K . compAlgG') e) haltG haltG)


instance(HFunctor f, CompAlgG f g) => CompAlgG' f g where
    compAlgG' x h c = compAlgG (hfmap (K . ($ h) . unK) x) c

instance (EXCEPT :<: g) => CompAlgG' Except g where
  compAlgG' (Throw)   h _ = throwG h
  compAlgG' (Catch (K x) (K h')) h c = letx c (\v -> markG |> letx (h' h |> Var v) 
                                          (\vh -> x (Var vh) |> unmarkG |> Var v))


---------------------
-- Virtual Machine --
---------------------

instance Exec EXCEPT where
    execAlg (THROW c) (k, s) = unwind' c (k, s)
    execAlg (MARK c) (k, s) = c (HAN' : k, s)
    execAlg (UNMARK c) (e : HAN' : k, s) = c (e : k, s)

unwind' c (HAN' : k, s) = c (k ,s)
unwind' c (_ : k, s) = unwind' c (k, s)
unwind' c ([], s) = ([], s)
