{-# LANGUAGE TypeOperators, MultiParamTypeClasses, 
             FlexibleInstances, DeriveFunctor, FlexibleContexts,
             OverlappingInstances #-}


module Fixpoint where

-------------------------
-- Fixpoints and Folds --
-------------------------

data Fix f = In (f (Fix f))
type Language = Fix

fold :: Functor f => (f a -> a) -> Fix f -> a
fold f (In t) = f (fmap (fold f) t)

infixr 8 :+:
data (f :+: g) e = Inl (f e) | Inr (g e)      deriving (Functor)

instance (Show a, Show (f a), Show (g a)) => Show ((f :+: g) a) where
    show (Inl x) = show x
    show (Inr x) = show x


class sub :<: sup where
  inj :: sub a -> sup a
  prj :: sup a -> Maybe (sub a)
  
instance (:<:) f f where
  inj = id
  prj = Just
  
instance (:<:) f (f :+: g) where
  inj = Inl
  prj (Inl x) = Just x
  prj (Inr _) = Nothing
  
instance ((:<:) f h) => (:<:) f (g :+: h) where
  inj = Inr . inj
  prj (Inr x) = prj x
  prj (Inl _) = Nothing
  
inject :: (g :<: f) => g (Language f) -> Language f
inject = In . inj
