{-# LANGUAGE TypeOperators, 
             FlexibleContexts,
             DeriveFunctor,
             NoMonomorphismRestriction #-}


-----------------------------------
-- Syntax of the Target Language --
-----------------------------------

module TargetLanguage where


import SourceLanguage (Ref)
import Graph
import Fixpoint
import Prelude hiding (EQ, LT, rem)



data ARITH e = PUSH Int e | ADD e | MUL e   deriving (Functor, Show)

data COND e = JPC e e  deriving (Functor, Show)

data HALT e = HALT   deriving (Functor, Show)

data EXCEPT e = THROW | UNMARK e | MARK e e deriving (Functor, Show)

data STATE e = GET Ref e | SET Ref e  deriving (Functor, Show)

data BOOL e = PUSHB Bool e | AND e | OR e  deriving (Functor, Show)

data COMP e = EQ e | LT e  deriving (Functor, Show)
  
pushG n c = injectG (PUSH n c)
addG    c = injectG (ADD c)
mulG    c = injectG (MUL c)
markG   = \h c -> injectG (MARK h c)
unmarkG = \  c -> injectG (UNMARK c)
throwG  = injectG (THROW)
jpcG  = \ l c -> injectG (JPC l c)
getG v c = injectG (GET v c)
setG v c = injectG (SET v c) 
haltG = injectG HALT

pushbG n c = injectG (PUSHB n c)
andG c = injectG (AND c)
orG c = injectG (OR c)
eqG    c = injectG (EQ c)
ltG    c = injectG (LT c)

-- Pretty printing instances

instance (Pretty f, Pretty g) => Pretty (f :+: g) where
    pretty (Inl x) = Inl $ pretty x
    pretty (Inr x) = Inr $ pretty x

instance Pretty COND where
    pretty (JPC x y) = JPC (paren x) (semi y)

instance Pretty EXCEPT where
    pretty (MARK x y) = MARK (paren x) (semi y)
    pretty x = simpl x

instance Pretty ARITH where
instance Pretty HALT where
instance Pretty STATE where
instance Pretty BOOL where
instance Pretty COMP where

